/**********************************************************************/
/*                                                                    */
/*   S O K O B A N  1.3 - main.cpp                                    */
/*                                                                    */
/*--------------------------------------------------------------------*/
/*                                 Arno Puder  07/92                  */
/*                                 Hohner Str. 16                     */
/*                                 53819 Neunkirchen                  */
/*                                 ---------------------------------  */
/*                          email: puder@informatik.uni-frankfurt.de  */
/**********************************************************************/


//#define SHAREWARE

#include <owl.h>
#include <bwcc.h>
#pragma hdrstop
#include <stdlib.h>
#include <string.h>
#include "sokoban.h"



/*
 * Game Field
 */

TGameField::TGameField( int		ALevel,
						HANDLE	AHBmpWall,
						HANDLE	AHBmpBall,
						HANDLE	AHBmpMan,
						HANDLE	AHBmpGoal,
						HANDLE	AHBmpSpace,
						HANDLE	AHBmpManGoal,
						HANDLE	AHBmpBallGoal )
{
	int			fh;
	OFSTRUCT	of;
	char		Name[ 30 ];
	BITMAP		bm;
	char		Char;
	int			NumBall, NumGoal;
	int			x, y;

	Level			= ALevel;
	Moves			= 0;
	HBmpWall  		= AHBmpWall;
	HBmpBall  		= AHBmpBall;
	HBmpMan   		= AHBmpMan;
	HBmpGoal  		= AHBmpGoal;
	HBmpSpace 		= AHBmpSpace;
	HBmpManGoal		= AHBmpManGoal;
	HBmpBallGoal	= AHBmpBallGoal;
	GetObject( HBmpWall, sizeof( BITMAP ), (LPSTR) &bm );
	XBmpSize = bm.bmWidth;
	YBmpSize = bm.bmHeight;
	wsprintf( Name, "level.%d", Level );
	fh = OpenFile( Name, &of, OF_READ );
	if( fh == -1 ) {
		Error = ERR_NOT_AVAIL;
		return;
	}
	for( x = 0; x < MAX_SCREEN_X; x++ )
		for( y = 0; y < MAX_SCREEN_Y; y++ )
			Screen[ x ][ y ] = CHAR_SPACE;
	x = y = 0;
	SizeX = SizeY = 0;
	ManX = ManY = 0;
	NumBall = NumGoal = 0;
	while( _lread( fh, (LPSTR) &Char, 1 ) == 1 ) {
		switch( Char ) {
			case CHAR_MAN:
				ManX = x;
				ManY = y;
				// Fallthrough
			case CHAR_WALL:
			case CHAR_BALL:
			case CHAR_GOAL:
			case CHAR_BALLGOAL:
			case CHAR_SPACE:
				Screen[ x++ ][ y ] = Char;
				if( x == MAX_SCREEN_X ) {
					Error = ERR_TOO_BIG;
					_lclose( fh );
					return;
				}
				if( Char == CHAR_BALL ) NumBall++;
				if( Char == CHAR_GOAL ) NumGoal++;
				break;
			case NEW_LINE:
				SizeX = ( SizeX > x ) ? SizeX : x;
				y++;
				if( y == MAX_SCREEN_Y ) {
					Error = ERR_TOO_BIG;
					_lclose( fh );
					return;
				}
				x = 0;
				break;
			case FORM_FEED:
				break;
			default:
				Error = ERR_BAD_SCREEN;
				_lclose( fh );
				return;
		}
	}
	_lclose( fh );
	if( ManX == 0 || ManY == 0 || NumBall != NumGoal ) {
		Error = ERR_BAD_SCREEN;
		return;
	}
	SizeY = y + 1;
	if( x == 0 ) SizeY--;
	LastPiece = CHAR_SPACE;
	Error = NO_ERR;
}




void TGameField::DrawPiece( HDC hdc, int X, int Y )
{
	HDC		hdcMem;
	BITMAP	bm;
	HANDLE	HBitmap;
	POINT	ptSize, ptOrg;

	switch( Screen[ X ][ Y ] ) {
		case CHAR_WALL:
			HBitmap = HBmpWall;
			break;
		case CHAR_BALL:
			HBitmap = HBmpBall;
			break;
		case CHAR_MAN:
			HBitmap = HBmpMan;
			break;
		case CHAR_GOAL:
			HBitmap = HBmpGoal;
			break;
		case CHAR_SPACE:
			HBitmap = HBmpSpace;
			break;
		case CHAR_MANGOAL:
			HBitmap = HBmpManGoal;
			break;
		case CHAR_BALLGOAL:
			HBitmap = HBmpBallGoal;
			break;
		default:
			return;
	}
	hdcMem = CreateCompatibleDC( hdc );
	SelectObject( hdcMem, HBitmap );
	SetMapMode( hdcMem, GetMapMode( hdc ) );
	GetObject( HBitmap, sizeof( BITMAP ), (LPSTR) &bm );
	ptSize.x = bm.bmWidth;
	ptSize.y = bm.bmHeight;
	DPtoLP( hdc, &ptSize, 1 );
	ptOrg.x = 0;
	ptOrg.y = 0;
	DPtoLP( hdcMem, &ptOrg, 1 );
	BitBlt( hdc, XBmpPos + X * XBmpSize, YBmpPos + Y * YBmpSize,
			ptSize.x, ptSize.y, hdcMem, ptOrg.x, ptOrg.y, SRCCOPY );
	DeleteDC( hdcMem );
}



void TGameField::DrawField( HDC hdc, int Width, int Height )
{
	int X, Y;

	XBmpPos = ( Width - XBmpSize * SizeX ) / 2;
	if( XBmpPos < XBmpSize ) XBmpPos = XBmpSize;
	YBmpPos = ( Height - YBmpSize * SizeY ) / 2;
	if( YBmpPos < YBmpSize ) YBmpPos = YBmpSize;
	for( X = 0; X < SizeX; X++ )
		for( Y = 0; Y < SizeY; Y++ )
			DrawPiece( hdc, X, Y );
}



void TGameField::Move( HDC hdc, int DeltaX, int DeltaY )
{
	char FirstNext;
	char SecondNext;

	FirstNext = Screen[ ManX + DeltaX ][ ManY + DeltaY ];
	switch( FirstNext ) {
		case CHAR_SPACE:
			Screen[ ManX ][ ManY ] = LastPiece;
			LastPiece = CHAR_SPACE;
			Screen[ ManX + DeltaX ][ ManY + DeltaY ] = CHAR_MAN;
			DrawPiece( hdc, ManX, ManY );
			DrawPiece( hdc, ManX + DeltaX, ManY + DeltaY );
			ManX += DeltaX;
			ManY += DeltaY;
			Moves++;
			break;
		case CHAR_GOAL:
			Screen[ ManX ][ ManY ] = LastPiece;
			LastPiece = CHAR_GOAL;
			Screen[ ManX + DeltaX ][ ManY + DeltaY ] = CHAR_MANGOAL;
			DrawPiece( hdc, ManX, ManY );
			DrawPiece( hdc, ManX + DeltaX, ManY + DeltaY );
			ManX += DeltaX;
			ManY += DeltaY;
			Moves++;
			break;
		case CHAR_BALLGOAL:
		case CHAR_BALL:
			SecondNext = Screen[ ManX + DeltaX + DeltaX ]
							   [ ManY + DeltaY + DeltaY ];
			switch( SecondNext ) {
				case CHAR_SPACE:
					Screen[ ManX + DeltaX + DeltaX ]
						  [ ManY + DeltaY + DeltaY ] = CHAR_BALL;
					Screen[ ManX + DeltaX ][ ManY + DeltaY ] =
							( FirstNext == CHAR_BALL ) ? CHAR_MAN : CHAR_MANGOAL;
					Screen[ ManX ][ ManY ] = LastPiece;
					LastPiece = ( FirstNext == CHAR_BALL ) ? CHAR_SPACE : CHAR_GOAL;
					DrawPiece( hdc, ManX, ManY );
					DrawPiece( hdc, ManX + DeltaX, ManY + DeltaY );
					DrawPiece( hdc, ManX + DeltaX + DeltaX,
									ManY + DeltaY + DeltaY );
					ManX += DeltaX;
					ManY += DeltaY;
					Moves++;
					break;
				case CHAR_GOAL:
					Screen[ ManX + DeltaX + DeltaX ]
						  [ ManY + DeltaY + DeltaY ] = CHAR_BALLGOAL;
					Screen[ ManX + DeltaX ][ ManY + DeltaY ] =
							( FirstNext == CHAR_BALL ) ? CHAR_MAN : CHAR_MANGOAL;
					Screen[ ManX ][ ManY ] = LastPiece;
					LastPiece = ( FirstNext == CHAR_BALL ) ? CHAR_SPACE : CHAR_GOAL;
					DrawPiece( hdc, ManX, ManY );
					DrawPiece( hdc, ManX + DeltaX, ManY + DeltaY );
					DrawPiece( hdc, ManX + DeltaX + DeltaX,
									ManY + DeltaY + DeltaY );
					ManX += DeltaX;
					ManY += DeltaY;
					Moves++;
					break;
			}
			break;
	}
}



void TGameField::MoveUp( HDC hdc )
{
	Move( hdc, 0, -1 );
}



void TGameField::MoveDown( HDC hdc )
{
	Move( hdc, 0, 1 );
}



void TGameField::MoveLeft( HDC hdc )
{
	Move( hdc, -1, 0 );
}




void TGameField::MoveRight( HDC hdc )
{
	Move( hdc, 1, 0 );
}




BOOL TGameField::ScreenFinished()
{
	int X, Y;

	for( X = 0; X < SizeX; X++ )
		for( Y = 0; Y < SizeY; Y++ )
			if( Screen[ X ][ Y ] == CHAR_BALL )
				return FALSE;
	return TRUE;
}



/*
 * New Level input dialog
 */

TLevelDialog::TLevelDialog( PTWindowsObject AParent, int AId, int ALevel, int *ANewLevel ) :
	TDialog( AParent, AId )
{
	Level = ALevel;
	NewLevel = ANewLevel;
	CurrLevel = new TStatic( this, IDD_LEVEL, 20 );
	EditField = new TEdit( this, IDD_EDIT_LEVEL, 0 );
	EditField->DisableTransfer();
}



void TLevelDialog::SetupWindow()
{
	char Buf[ 21 ];

	TDialog::SetupWindow();
	if( Level == 0 )
		CurrLevel->SetText( "-" );
	else {
		wsprintf( Buf, "%d", Level );
		CurrLevel->SetText( (LPSTR) Buf );
	}
}




BOOL TLevelDialog::CanClose()
{
	char Buf[ 20 ];

	EditField->GetText( Buf, 20 );
	if( ( *NewLevel = atoi( Buf ) ) == 0 ) {
		MessageBeep( 0 );
		return FALSE;
	} else
		return TRUE;
}





/*
 * Main Window Object
 */


TMainWindow::TMainWindow( PTWindowsObject AParent, LPSTR ATitle ) :
				TBWindow( AParent, ATitle )
{
	HBmpWall  = LoadBitmap( GetModule()->hInstance,
							MAKEINTRESOURCE( BMP_WALL ) );
	HBmpBall  = LoadBitmap( GetModule()->hInstance,
							MAKEINTRESOURCE( BMP_BALL ) );
	HBmpMan   = LoadBitmap( GetModule()->hInstance,
							MAKEINTRESOURCE( BMP_MAN ) );
	HBmpGoal  = LoadBitmap( GetModule()->hInstance,
							MAKEINTRESOURCE( BMP_GOAL ) );
	HBmpSpace = LoadBitmap( GetModule()->hInstance,
							MAKEINTRESOURCE( BMP_SPACE ) );
	HBmpManGoal = LoadBitmap( GetModule()->hInstance,
							MAKEINTRESOURCE( BMP_MANGOAL ) );
	HBmpBallGoal = LoadBitmap( GetModule()->hInstance,
							MAKEINTRESOURCE( BMP_BALLGOAL ) );
	HMainMenuEng = LoadMenu( GetModule()->hInstance,
							MAKEINTRESOURCE( IDM_MAIN_MENU ) );
	HMainMenuGer = LoadMenu( GetModule()->hInstance,
							MAKEINTRESOURCE( IDM_MAIN_MENU_GER ) );
	GameField = NULL;
	SavedGame = new TGameField;
}



TMainWindow::~TMainWindow()
{
	char Buf[ 20 ];

	KillTimer( HWindow, 1 );
	if( GameField )
		delete GameField;
	delete SavedGame;
	wsprintf( Buf, "%d", Level );
	WritePrivateProfileString( "Sokoban", "Level", Buf, "SOKOBAN.INI" );
	WritePrivateProfileString( "Sokoban", "Language",
							   Language == ENGLISH ? "English" : "German",
							   "SOKOBAN.INI" );
	WritePrivateProfileString( "Sokoban", "ShowTime",
							   ShowTime ? "Yes" : "No", "SOKOBAN.INI" );
	wsprintf( Buf, "%d", Time );
	WritePrivateProfileString( "Sokoban", "Time", Buf, "SOKOBAN.INI" );
	wsprintf( Buf, "%d", Moves );
	WritePrivateProfileString( "Sokoban", "Moves", Buf, "SOKOBAN.INI" );
	WritePrivateProfileString( "Sokoban", "ShowMoves",
							   ShowMoves ? "Yes" : "No", "SOKOBAN.INI" );
	DeleteObject( HBmpWall );
	DeleteObject( HBmpBall );
	DeleteObject( HBmpMan );
	DeleteObject( HBmpGoal );
	DeleteObject( HBmpSpace );
	DeleteObject( HBmpManGoal );
	DeleteObject( HBmpBallGoal );
}



void TMainWindow::SetupWindow()
{
	char Buf[ 20 ];

	TWindow::SetupWindow();

	// Get desired language
	GetPrivateProfileString( "Sokoban", "Language", "English", Buf, 20, "SOKOBAN.INI" );
	if( strcmp( Buf, "English" ) == 0 )
		Language = ENGLISH;
	else
		Language = GERMAN;

	// Display time?
	GetPrivateProfileString( "Sokoban", "ShowTime", "No", Buf, 20, "SOKOBAN.INI" );
	ShowTime = strcmp( Buf, "Yes" ) == 0;

	// Get time
	Time = GetPrivateProfileInt( "Sokoban", "Time", 0, "SOKOBAN.INI" );

	// Display # of moves?
	GetPrivateProfileString( "Sokoban", "ShowMoves", "Yes", Buf, 20, "SOKOBAN.INI" );
	ShowMoves = strcmp( Buf, "Yes" ) == 0;

	// Get moves
	Moves = GetPrivateProfileInt( "Sokoban", "Moves", 0, "SOKOBAN.INI" );

	// Load menu
	SetMenu( HWindow, Language == ENGLISH ? HMainMenuEng : HMainMenuGer );
	CheckMenuItem( GetMenu( HWindow ), CM_SHOW_TIME,
					ShowTime ? MF_CHECKED : MF_UNCHECKED );
	CheckMenuItem( GetMenu( HWindow ), CM_SHOW_MOVES,
					ShowMoves ? MF_CHECKED : MF_UNCHECKED );
#ifndef SHAREWARE
	RemoveMenu( GetMenu( HWindow ), CM_REGISTER, MF_BYCOMMAND );
#endif

	// Load level
	NewGame( GetPrivateProfileInt( "Sokoban", "Level", 1, "SOKOBAN.INI" ) );

	// Set timer
	SetTimer( HWindow, 1, 1000, NULL );
}



void TMainWindow::Paint( HDC hdc, PAINTSTRUCT _FAR &ps )
{
	TEXTMETRIC	tm;
	RECT		Rect;

	if( GameField ) {
		GetTextMetrics( hdc, &tm );
		ShowStat( hdc );
		if( tm.tmExternalLeading + tm.tmHeight < ps.rcPaint.bottom ) {
			GetClientRect( HWindow, &Rect );
			GameField->DrawField( hdc, Rect.right, Rect.bottom );
		}
	}
}



void TMainWindow::ShowStat( HDC hdc )
{
	char		Buf[ 30 ];
	TEXTMETRIC	tm;
	int			X;
	int			CurrMoves;
	WORD		Time1;
	WORD		Hours, Min, Sec;

	SetBkColor( hdc, RGB( 192, 192, 192 ) );
	GetTextMetrics( hdc, &tm );
	X = tm.tmAveCharWidth;
	if( ShowMoves ) {
		CurrMoves = Moves + GameField->GetNumMoves();
		if( Language == ENGLISH )
			wsprintf( Buf, "Moves: %d", CurrMoves );
		else
			wsprintf( Buf, "Zge: %d", CurrMoves );
		TextOut( hdc, X, tm.tmExternalLeading, Buf, strlen( Buf ) );
		X += 15 * tm.tmAveCharWidth;
	}
	if( ShowTime ) {
		Hours = Time / ( 60 * 60 );
		Time1 = Time % ( 60 * 60 );
		Min = Time1 / 60;
		Sec = Time1 % 60;
		wsprintf( Buf, LANG( "Time: %2d:%02d:%02d", "Zeit: %2d:%02d:%02d" ),
					   Hours, Min, Sec );
		TextOut( hdc, X, tm.tmExternalLeading, Buf, strlen( Buf ) );
	}
}



void TMainWindow::UpdateStat( void )
{
	RECT		Rect;
	TEXTMETRIC	tm;
	HDC			hdc;

	GetClientRect( HWindow, &Rect );
	hdc = GetDC( HWindow );
	GetTextMetrics( hdc, &tm );
	ReleaseDC( HWindow, hdc );
	Rect.bottom = tm.tmExternalLeading + tm.tmHeight;
	InvalidateRect( HWindow, &Rect, TRUE );
}



void TMainWindow::NewGame( int ALevel )
{
	char *ErrMsg;
	char Caption[ 20 ];

#ifdef SHAREWARE
	if( Language == ENGLISH )
		BWCCMessageBox( HWindow, "Don't forget to register!", "Shareware SOKOBAN",
						MB_OK | MB_ICONHAND );
    else
		BWCCMessageBox( HWindow, "Vergessen Sie nicht SOKOBAN zu registrieren!",
						"Shareware SOKOBAN", MB_OK | MB_ICONHAND );
#endif
	if( GameField ) delete GameField;
	GameField = NULL;
	Level = ALevel;
	EnableMenuItem( GetMenu( HWindow ), CM_LAST_MOVE, MF_GRAYED );
	if( Level == 0 ) return;
	GameField = new TGameField( ALevel,
								HBmpWall,
								HBmpBall,
								HBmpMan,
								HBmpGoal,
								HBmpSpace,
								HBmpManGoal,
								HBmpBallGoal );
	switch( GameField->GetStatus() ) {
		case NO_ERR:
			Level = ALevel;
#ifdef SHAREWARE
			wsprintf( Caption, "Shareware Sokoban - %d", Level );
#else
			wsprintf( Caption, "Sokoban - %d", Level );
#endif
			SetCaption( Caption );
			EnableMenuItem( GetMenu( HWindow ), CM_RESTORE, MF_ENABLED );
			InvalidateRect( HWindow, NULL, TRUE );
			return;
		case ERR_TOO_BIG:
			ErrMsg = LANG( "Screen description is too big.",
						   "Spielbeschreibung ist zu gro." );
			break;
		case ERR_NOT_AVAIL:
			ErrMsg = LANG( "Couldn't find screen description for this level.",
						   "Keine Spielbeschreibung fr diese Ebene gefunden." );
			break;
		case ERR_BAD_SCREEN:
			ErrMsg = LANG( "Screen file has bad format.",
						   "Spielbeschreibung hat falsches Format." );
			break;
	}
	delete GameField;
	GameField = NULL;
	Level = 0;
	InvalidateRect( HWindow, NULL, TRUE );
	BWCCMessageBox( HWindow, ErrMsg, LANG( "Error", "Fehler" ), MB_ICONSTOP | MB_OK );
#ifdef SHAREWARE
	SetCaption( "Shareware Sokoban" );
#else
	SetCaption( "Sokoban" );
#endif
	EnableMenuItem( GetMenu( HWindow ), CM_RESTORE, MF_GRAYED );
}



void TMainWindow::Snapshot()
{
	*SavedGame = *GameField;
	EnableMenuItem( GetMenu( HWindow ), CM_LAST_MOVE, MF_ENABLED );
}



void TMainWindow::CMLoadFirst( RTMessage )
{
	Moves = 0;
	Time = 0;
	NewGame( 1 );
}



void TMainWindow::CMLoadArb( RTMessage )
{
	PTLevelDialog	LevelDialog;
	int				NewLevel;

	LevelDialog = new TLevelDialog( this,
								LANG( DLG_NEW_LEVEL, DLG_NEW_LEVEL_GER ),
								Level, &NewLevel );
	if( ( GetApplication()->ExecDialog( LevelDialog ) ) == IDOK )
		NewGame( NewLevel );
}



void TMainWindow::CMRestore( RTMessage )
{
	NewGame( Level );
}



void TMainWindow::CMLastMove( RTMessage )
{
	*GameField = *SavedGame;
	EnableMenuItem( GetMenu( HWindow ), CM_LAST_MOVE, MF_GRAYED );
	InvalidateRect( HWindow, NULL, FALSE );
}



void TMainWindow::CMShowTime( RTMessage )
{
	ShowTime = !ShowTime;
	CheckMenuItem( GetMenu( HWindow ), CM_SHOW_TIME,
					ShowTime ? MF_CHECKED : MF_UNCHECKED );
	UpdateStat();
}




void TMainWindow::CMShowMoves( RTMessage )
{
	ShowMoves = !ShowMoves;
	CheckMenuItem( GetMenu( HWindow ), CM_SHOW_MOVES,
					ShowMoves ? MF_CHECKED : MF_UNCHECKED );
	UpdateStat();
}



void TMainWindow::CMLanguage( RTMessage )
{
	if( Language == ENGLISH ) {
		SetMenu( HWindow, HMainMenuGer );
		Language = GERMAN;
	} else {
		SetMenu( HWindow, HMainMenuEng );
		Language = ENGLISH;
	}
#ifndef SHAREWARE
	RemoveMenu( GetMenu( HWindow ), CM_REGISTER, MF_BYCOMMAND );
#endif
	CheckMenuItem( GetMenu( HWindow ), CM_SHOW_TIME,
					ShowTime ? MF_CHECKED : MF_UNCHECKED );
	CheckMenuItem( GetMenu( HWindow ), CM_SHOW_MOVES,
					ShowMoves ? MF_CHECKED : MF_UNCHECKED );
	EnableMenuItem( GetMenu( HWindow ), CM_LAST_MOVE, MF_GRAYED );
	if( Level == 0 )
		EnableMenuItem( GetMenu( HWindow ), CM_RESTORE, MF_GRAYED );
	else
		UpdateStat();
}



void TMainWindow::CMInfo( RTMessage )
{
	GetApplication()->ExecDialog(
			new TDialog( this, LANG( DLG_INFO, DLG_INFO_GER ) ) );
}



void TMainWindow::CMRegister( RTMessage )
{
	GetApplication()->ExecDialog(
			new TDialog( this, LANG( DLG_REGISTER_ENG, DLG_REGISTER_GER ) ) );
}



void TMainWindow::CMAbout( RTMessage )
{
	GetApplication()->ExecDialog( new TDialog( this, DLG_ABOUT ) );
}



void TMainWindow::WMKeyDown( RTMessage Msg )
{
	HDC	hdc;

	if( !GameField ) return;
	Snapshot();
	hdc = GetDC( HWindow );
	switch( Msg.WParam ) {
		case VK_UP:
			GameField->MoveUp( hdc );
			break;
		case VK_DOWN:
			GameField->MoveDown( hdc );
			break;
		case VK_RIGHT:
			GameField->MoveRight( hdc );
			break;
		case VK_LEFT:
			GameField->MoveLeft( hdc );
			break;
	}
	ShowStat( hdc );
	ReleaseDC( HWindow, hdc );
	if( GameField->ScreenFinished() ) {
		Moves += GameField->GetNumMoves();
		NewGame( ++Level );
	}
}



void TMainWindow::WMTimer( RTMessage )
{
	if( GameField ) {
		Time++;
		UpdateStat();
	}
}



/*
 * Application Object
 */

void TSokobanApp::InitInstance()
{
	TApplication::InitInstance();
	HAccTable = LoadAccelerators( this->hInstance, MAKEINTRESOURCE( ACC_SOKOBAN ) );
}



void TSokobanApp::InitMainWindow()
{
	BWCCGetVersion();
	MainWindow = new TMainWindow( NULL, Name );
}



int PASCAL WinMain( HANDLE hInstance, HANDLE hPrevInstance,
					LPSTR lpCmdLine, int nCmdShow )
{
	TSokobanApp App( "Shareware Sokoban", hInstance, hPrevInstance,
						lpCmdLine, nCmdShow );

	App.Run();
	return App.Status;
}
