/****************************************************************/
/*                                                              */
/*   S O K O B A N  1.0 - sokoban.h                             */
/*                                                              */
/*--------------------------------------------------------------*/
/*                                        Arno Puder  07/92     */
/*                                        Hohner Str. 16        */
/*                                        5206 Neunkirchen 1    */
/*                                        --------------------  */
/*                                 email: puder@rhrk.uni-kl.de  */
/****************************************************************/

#ifndef __SOKOBAN_H__
#define __SOKOBAN_H__

#include <bwindow.h>
#include <static.h>
#include <edit.h>
#include "resource.h"


#define ENGLISH			1
#define GERMAN			2

// Language macro
#define LANG( e, g )	( Language == ENGLISH ? (e) : (g) )

#define NO_ERR			0
#define ERR_TOO_BIG		-1
#define ERR_NOT_AVAIL	-2
#define ERR_BAD_SCREEN	-3

#define MAX_SCREEN_X	25
#define MAX_SCREEN_Y	25


#define CHAR_WALL		'#'
#define CHAR_BALL		'$'
#define CHAR_MAN		'@'
#define CHAR_GOAL		'.'
#define CHAR_SPACE  	' '
#define CHAR_MANGOAL	'!'
#define CHAR_BALLGOAL	'*'
#define NEW_LINE		10
#define FORM_FEED		13


_CLASSDEF( TGameField )
class TGameField {
	private:
		int		Error;
		HANDLE	HBmpWall;
		HANDLE	HBmpBall;
		HANDLE	HBmpMan;
		HANDLE	HBmpGoal;
		HANDLE	HBmpSpace;
		HANDLE	HBmpManGoal;
		HANDLE	HBmpBallGoal;
		int		Level;
		int		ManX, ManY;
		int		SizeX, SizeY;
		char	Screen[ MAX_SCREEN_X ][ MAX_SCREEN_Y ];
		char	LastPiece;
		int		XBmpPos, YBmpPos;
		int		XBmpSize, YBmpSize;
		int		Moves;
	public:
		TGameField( int		ALevel,
					HANDLE	AHBmpWall,
					HANDLE	AHBmpBall,
					HANDLE	AHBmpMan,
					HANDLE	AHBmpGoal,
					HANDLE	AHBmpSpace,
					HANDLE	AHBmpManGoal,
					HANDLE	AHBmpBallGoal );
		TGameField( void ) {};
		int GetStatus( void ) { return Error; };
		int GetNumMoves( void ) { return Moves; };
		void DrawPiece( HDC hdc, int X, int Y );
		void DrawField( HDC hdc, int Width, int Height );
		void Move( HDC hdc, int DeltaX, int DeltaY );
		void MoveUp( HDC hdc );
		void MoveDown( HDC hdc );
		void MoveRight( HDC hdc );
		void MoveLeft( HDC hdc );
		BOOL ScreenFinished();
};



_CLASSDEF( TLevelDialog )
class TLevelDialog : public TDialog {
	private:
		int			Level;
		int			*NewLevel;
		PTStatic	CurrLevel;
		PTEdit		EditField;
	public:
		TLevelDialog( PTWindowsObject AParent, int AId, int ALevel, int *ANewLevel );
		virtual void SetupWindow();
		virtual BOOL CanClose();
};




_CLASSDEF( TMainWindow )
class TMainWindow : public TBWindow {
	private:
		PTGameField	GameField;
		PTGameField	SavedGame;
		HANDLE		HBmpWall;
		HANDLE		HBmpBall;
		HANDLE		HBmpMan;
		HANDLE		HBmpGoal;
		HANDLE		HBmpSpace;
		HANDLE		HBmpManGoal;
		HANDLE		HBmpBallGoal;
		HANDLE		HMainMenuEng;
		HANDLE		HMainMenuGer;
		int			Level;
		int			Language;
		BOOL		ShowTime;
		WORD		Time;
		BOOL		ShowMoves;
		WORD		Moves;
	public:
		TMainWindow( PTWindowsObject AParent, LPSTR ATitle );
		~TMainWindow();
		virtual void GetWindowClass( WNDCLASS &AWndClass )
			{
				TBWindow::GetWindowClass( AWndClass );
				AWndClass.hIcon = LoadIcon( GetModule()->hInstance,
									MAKEINTRESOURCE( ICO_SOKOBAN ) );
			};
		virtual void SetupWindow();
		virtual void Paint( HDC hdc, PAINTSTRUCT _FAR &PaintInfo );
		void ShowStat( HDC hdc );
		void UpdateStat( void );
		void NewGame( int ALevel );
		void Snapshot( void );
		virtual void CMLoadFirst( RTMessage ) =
				[ CM_FIRST + CM_LOAD_FIRST ];
		virtual void CMLoadArb( RTMessage ) =
				[ CM_FIRST + CM_LOAD_ARB ];
		virtual void CMRestore( RTMessage ) =
				[ CM_FIRST + CM_RESTORE ];
		virtual void CMLastMove( RTMessage ) =
				[ CM_FIRST + CM_LAST_MOVE ];
		virtual void CMShowTime( RTMessage ) =
				[ CM_FIRST + CM_SHOW_TIME ];
		virtual void CMShowMoves( RTMessage ) =
				[ CM_FIRST + CM_SHOW_MOVES ];
		virtual void CMLanguage( RTMessage ) =
				[ CM_FIRST + CM_LANGUAGE ];
		virtual void CMInfo( RTMessage ) =
				[ CM_FIRST + CM_INFO ];
		virtual void CMRegister( RTMessage ) =
				[ CM_FIRST + CM_REGISTER ];
		virtual void CMAbout( RTMessage ) =
				[ CM_FIRST + CM_ABOUT ];
		virtual void WMKeyDown( RTMessage ) =
				[ WM_FIRST + WM_KEYDOWN ];
		virtual void WMTimer( RTMessage ) =
				[ WM_FIRST + WM_TIMER ];
};



class TSokobanApp : public TApplication {
	public:
		TSokobanApp( LPSTR AName, HANDLE hInstance, HANDLE hPrevInstance,
					  LPSTR lpCmdLine, int nCmdShow ) :
			TApplication( AName, hInstance, hPrevInstance, lpCmdLine, nCmdShow ) {};
		virtual void InitInstance();
		virtual void InitMainWindow();
};



#endif